/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "xsolidp.h"
#include "icon_bitmap"

/***************************************************************************
  Declare external variables
****************************************************************************/

extern descript ws_des;
extern robot therobot;
extern FILE *log_fd;

/***************************************************************************
  Declare global variables
****************************************************************************/

Display *mydisp;
int screen;

/***************************************************************************
  Function to display path
****************************************************************************/

void visuel(path X,segment *world_seg,int world_nbseg,double f,
	    double d,double theta,double phi,double ty,double tz) {

  int argc = 0;
  char **argv = 0;
  Window mywin;
  unsigned int width, height;
  unsigned int display_width, display_height;
  int x=0, y=0;
  unsigned int border_width = 4;
  char *window_name = "STANFORD UNIVERSITY GENERAL PATH PLANNER (G.P.P.)";
  char *icon_name = "G.P.P.";
  Pixmap icon_pixmap;
  XSizeHints size_hints;
  XEvent report;
  XFontStruct *font_info;
  GC gc;
  char *display_name = NULL;
  char buffer[20];
  int buffsize = 20;
  KeySym keysym;
  XComposeStatus compose;
  int i=0,j,iold;
  segment **robot_seg;
  int robot_nbseg;
  segment2D *world_seg2d,*robot_seg2d;
  int charcount;
  
  robot_nbseg = therobot.nblink;
  robot_seg   = (segment **)calloc2D(X.dim,robot_nbseg,sizeof(segment));
  world_seg2d = (segment2D *)calloc(world_nbseg,sizeof(segment2D));
  robot_seg2d = (segment2D *)calloc(robot_nbseg,sizeof(segment2D));

  for(j = 0;j < X.dim;++j)
    compute_robot_seg(X.d[j],robot_seg[j]);
  
  if( ( mydisp = XOpenDisplay(display_name)) == NULL) {
    (void) fprintf(stderr,"Robotvis: cannot connect to X server %s\n",
		   XDisplayName(display_name)); 
    exit(-1);
  }
  
  screen = DefaultScreen(mydisp);
  display_width = DisplayWidth(mydisp, screen);
  display_height = DisplayHeight(mydisp, screen);
  width = 256*3;
  height = 256*3;
  mywin = XCreateSimpleWindow(mydisp, RootWindow(mydisp,screen), 
				  x, y, width, height, border_width, 
				  BlackPixel(mydisp,screen),
				  WhitePixel(mydisp,screen));

  icon_pixmap = XCreateBitmapFromData(mydisp, mywin, icon_bitmap_bits,
				      icon_bitmap_width, icon_bitmap_height);

  size_hints.flags = PPosition | PSize | PMinSize;
  size_hints.x = x;
  size_hints.y = y;
  size_hints.width = width;
  size_hints.height = height;
  size_hints.min_width = 350;
  size_hints.min_height = 250;
  
  XSetStandardProperties(mydisp, mywin, window_name, icon_name, 
			 icon_pixmap, argv, argc, &size_hints);
  
  XSelectInput(mydisp,mywin, ExposureMask | KeyPressMask |
	       ButtonPressMask |   StructureNotifyMask);
  
  load_font(&font_info);
  get_GC(mywin, &gc ,font_info );
  XMapWindow(mydisp,mywin);
  
  while(1) {
    XNextEvent( mydisp, &report);
    switch (report.type) {

    case Expose:
      /* Get rid of the other Expose on the queue */
      while(XCheckTypedEvent(mydisp, Expose, &report));
      
      /* Place graphics in window */
      perspective(world_seg,world_seg2d,world_nbseg,f,d,theta,phi,ty,tz);
      draw_seg(mywin, gc, width, height,world_seg2d,world_nbseg);
      perspective(robot_seg[i],robot_seg2d,robot_nbseg,f,d,theta,phi,ty,tz);
      draw_seg(mywin, gc, width, height,robot_seg2d,robot_nbseg);
      break;
      
    case ConfigureNotify:
      /* Window has been resized, change width and height to send 
	 to drawing routines in next Expose */
      width = report.xconfigure.width;
      height = report.xconfigure.height;
      i = 0;
      break;

    case ButtonPress:
      
      switch(report.xbutton.button){

      case Button1:
	XUnloadFont(mydisp,font_info->fid);		
	XFreeGC(mydisp, gc);
	XCloseDisplay(mydisp);
	exit(1);
	break;

      case Button2:
	
	theta += ( report.xbutton.x/(double)width - 0.5) ;
	phi   += ( report.xbutton.y/(double)height - 0.5) ;
	draw_seg(mywin, gc, width, height,world_seg2d,world_nbseg);
	perspective(world_seg,world_seg2d,world_nbseg,f,d,theta,phi,ty,tz);
	draw_seg(mywin, gc, width, height,world_seg2d,world_nbseg);
	draw_seg(mywin, gc, width, height,robot_seg2d,robot_nbseg);
	perspective(robot_seg[i],robot_seg2d,robot_nbseg,f,d,theta,phi,ty,tz);
	draw_seg(mywin, gc, width, height,robot_seg2d,robot_nbseg);
	break;

      case Button3:

	d += (report.xbutton.x/(double)width - 0.5);
	draw_seg(mywin, gc, width, height,world_seg2d,world_nbseg);
	perspective(world_seg,world_seg2d,world_nbseg,f,d,theta,phi,ty,tz);
	draw_seg(mywin, gc, width, height,world_seg2d,world_nbseg);
	draw_seg(mywin, gc, width, height,robot_seg2d,robot_nbseg);
	perspective(robot_seg[i],robot_seg2d,robot_nbseg,f,d,theta,phi,ty,tz);
	draw_seg(mywin, gc, width, height,robot_seg2d,robot_nbseg);
	break;

      } /* END SWITCH BUTTON */
      break;

    case KeyPress:
      charcount = XLookupString((XKeyEvent *)&report,buffer,buffsize,
				&keysym, &compose);
      if( (keysym != XK_Right) && (keysym != XK_Left) ) break;
      if(keysym == XK_Right ) {iold = i ; if(i < X.dim-1) i++;}
      if( keysym == XK_Left ) {iold = i ; if(i > 0) i--;}
      draw_seg(mywin, gc, width, height,robot_seg2d,robot_nbseg);
      perspective(robot_seg[i],robot_seg2d,robot_nbseg,f,d,theta,phi,ty,tz);
      draw_seg(mywin, gc, width, height,robot_seg2d,robot_nbseg);
      fprintf(stderr,"i = %d: (%4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f)\n",
	      i,X.d[i][0],X.d[i][1],X.d[i][2],X.d[i][3],X.d[i][4],
	      X.d[i][5],X.d[i][6]);
      break;

    default:
      /* All events selected by StructureNotifyMask except
	 ConfigureNotify are thrown away here since nothing is
	 done with them */
      break;
    }/* end switch */
  }/* end while */
}

/***************************************************************************
  function to define the graphic context 
****************************************************************************/

void get_GC(Window mywin, GC *gc, XFontStruct *font_info) {

  unsigned long valuemask = (GCFunction);
  XGCValues values;
  unsigned int line_width = 0;
  int line_style = LineSolid;
  int cap_style = CapRound;
  int join_style = JoinRound;
  values.function = GXinvert;
  
  
  *gc = XCreateGC(mydisp, mywin,valuemask, &values);
  XSetFont(mydisp, *gc, font_info->fid);
  XSetForeground(mydisp, *gc, BlackPixel(mydisp, screen));
  XSetLineAttributes(mydisp, *gc, line_width, line_style, cap_style, 
		     join_style);
}

/***************************************************************************
  Function to load fonts
****************************************************************************/

void load_font(XFontStruct **font_info){
  
  char *fontname = "courB14";
  
  if((*font_info = XLoadQueryFont(mydisp,fontname)) == NULL){
    (void) fprintf(stderr,"Cannot open courB14 font \n");
    exit(-1);
  }
}

/***************************************************************************
  Function to draw a 3d line segment
****************************************************************************/

void draw_seg(Window mywin,GC gc,unsigned int window_width, 
	      unsigned int window_height,segment2D *seg2d,int nbseg) {

  int i,au,av,bu,bv;
  
  for(i=0;i<nbseg;++i){
    au = (int)(seg2d[i].a.x * (window_height-1)); 
    av = (int)(seg2d[i].a.y * (window_width-1));
    bu = (int)(seg2d[i].b.x * (window_height-1));
    bv = (int)(seg2d[i].b.y * (window_width-1));
    XDrawLine(mydisp,mywin,gc,av,au,bv,bu);
  }
}

/***************************************************************************
  Function to compute the line segment for robot links
****************************************************************************/

void compute_robot_seg(configuration X,segment *seg) {

  register int i;
  fpoint3D kineX[MAXNBP];
  
  CFK(X,kineX);
  
  for(i=0;i<therobot.nblink;++i){
    seg[i].a.x = kineX[therobot.robotlink[i].endp1].x;
    seg[i].a.y = kineX[therobot.robotlink[i].endp1].y;
    seg[i].a.z = kineX[therobot.robotlink[i].endp1].z;
    
    seg[i].b.x = kineX[therobot.robotlink[i].endp2].x;
    seg[i].b.y = kineX[therobot.robotlink[i].endp2].y;
    seg[i].b.z = kineX[therobot.robotlink[i].endp2].z;
  }
#ifdef DEBUG
    fprintf(log_fd,"(%4.2f,%4.2f,%4.2f),(%4.2f,%4.2f,%4.2f),(%4.2f,%4.2f,%4.2f)\n",kineX[0].x,kineX[0].y,kineX[0].z,kineX[1].x,kineX[1].y,kineX[1].z,kineX[2].x,kineX[2].y,kineX[2].z);
#endif

}
